#![allow(clippy::many_single_char_names)]

use super::*;

/// A 4x4 column-major Matrix.
///
/// There's four columns, and each column is itself a `Vec4`.
#[derive(Clone, Copy, Default, PartialEq)]
#[repr(align(16), C)]
pub struct Mat4 {
  pub(crate) x_axis: Vec4,
  pub(crate) y_axis: Vec4,
  pub(crate) z_axis: Vec4,
  pub(crate) w_axis: Vec4,
}
unsafe impl Zeroable for Mat4 {}
unsafe impl Pod for Mat4 {}

impl core::fmt::Debug for Mat4 {
  /// Debug formats with each axis labeled and then shown as a 4-tuple in one
  /// long line.
  ///
  /// Passes the formatter along to the fields, so you can use any normal `f32`
  /// Debug format arguments that you like.
  fn fmt(&self, f: &mut core::fmt::Formatter) -> core::fmt::Result {
    f.write_str("Mat4 { x_axis: (")?;
    let [x, y, z, w]: [f32; 4] = cast(self.x_axis);
    core::fmt::Debug::fmt(&x, f)?;
    f.write_str(", ")?;
    core::fmt::Debug::fmt(&y, f)?;
    f.write_str(", ")?;
    core::fmt::Debug::fmt(&z, f)?;
    f.write_str(", ")?;
    core::fmt::Debug::fmt(&w, f)?;
    f.write_str("), y_axis: (")?;
    let [x, y, z, w]: [f32; 4] = cast(self.y_axis);
    core::fmt::Debug::fmt(&x, f)?;
    f.write_str(", ")?;
    core::fmt::Debug::fmt(&y, f)?;
    f.write_str(", ")?;
    core::fmt::Debug::fmt(&z, f)?;
    f.write_str(", ")?;
    core::fmt::Debug::fmt(&w, f)?;
    f.write_str("), z_axis: (")?;
    let [x, y, z, w]: [f32; 4] = cast(self.z_axis);
    core::fmt::Debug::fmt(&x, f)?;
    f.write_str(", ")?;
    core::fmt::Debug::fmt(&y, f)?;
    f.write_str(", ")?;
    core::fmt::Debug::fmt(&z, f)?;
    f.write_str(", ")?;
    core::fmt::Debug::fmt(&w, f)?;
    f.write_str("), w_axis: (")?;
    let [x, y, z, w]: [f32; 4] = cast(self.w_axis);
    core::fmt::Debug::fmt(&x, f)?;
    f.write_str(", ")?;
    core::fmt::Debug::fmt(&y, f)?;
    f.write_str(", ")?;
    core::fmt::Debug::fmt(&z, f)?;
    f.write_str(", ")?;
    core::fmt::Debug::fmt(&w, f)?;
    f.write_str(") }")
  }
}

impl core::fmt::Display for Mat4 {
  /// Display formats without labels in a 4 line style similar to standard math
  /// notation.
  ///
  /// Passes the formatter along to the fields, so you can use any normal `f32`
  /// Display format arguments that you like.
  fn fmt(&self, f: &mut core::fmt::Formatter) -> core::fmt::Result {
    let [x0, x1, x2, x3]: [f32; 4] = cast(self.x_axis);
    let [y0, y1, y2, y3]: [f32; 4] = cast(self.y_axis);
    let [z0, z1, z2, z3]: [f32; 4] = cast(self.z_axis);
    let [w0, w1, w2, w3]: [f32; 4] = cast(self.w_axis);
    //
    f.write_str("| ")?;
    core::fmt::Display::fmt(&x0, f)?;
    f.write_str(", ")?;
    core::fmt::Display::fmt(&y0, f)?;
    f.write_str(", ")?;
    core::fmt::Display::fmt(&z0, f)?;
    f.write_str(", ")?;
    core::fmt::Display::fmt(&w0, f)?;
    f.write_str(" |\n| ")?;
    core::fmt::Display::fmt(&x1, f)?;
    f.write_str(", ")?;
    core::fmt::Display::fmt(&y1, f)?;
    f.write_str(", ")?;
    core::fmt::Display::fmt(&z1, f)?;
    f.write_str(", ")?;
    core::fmt::Display::fmt(&w1, f)?;
    f.write_str(" |\n| ")?;
    core::fmt::Display::fmt(&x2, f)?;
    f.write_str(", ")?;
    core::fmt::Display::fmt(&y2, f)?;
    f.write_str(", ")?;
    core::fmt::Display::fmt(&z2, f)?;
    f.write_str(", ")?;
    core::fmt::Display::fmt(&w2, f)?;
    f.write_str(" |\n| ")?;
    core::fmt::Display::fmt(&x3, f)?;
    f.write_str(", ")?;
    core::fmt::Display::fmt(&y3, f)?;
    f.write_str(", ")?;
    core::fmt::Display::fmt(&z3, f)?;
    f.write_str(", ")?;
    core::fmt::Display::fmt(&w3, f)?;
    f.write_str(" |")
  }
}

impl core::fmt::LowerExp for Mat4 {
  /// LowerExp formats like Display, but with the lower exponent.
  ///
  /// Passes the formatter along to the fields, so you can use any normal `f32`
  /// LowerExp format arguments that you like.
  fn fmt(&self, f: &mut core::fmt::Formatter) -> core::fmt::Result {
    let [x0, x1, x2, x3]: [f32; 4] = cast(self.x_axis);
    let [y0, y1, y2, y3]: [f32; 4] = cast(self.y_axis);
    let [z0, z1, z2, z3]: [f32; 4] = cast(self.z_axis);
    let [w0, w1, w2, w3]: [f32; 4] = cast(self.w_axis);
    //
    f.write_str("| ")?;
    core::fmt::LowerExp::fmt(&x0, f)?;
    f.write_str(", ")?;
    core::fmt::LowerExp::fmt(&y0, f)?;
    f.write_str(", ")?;
    core::fmt::LowerExp::fmt(&z0, f)?;
    f.write_str(", ")?;
    core::fmt::LowerExp::fmt(&w0, f)?;
    f.write_str(" |\n| ")?;
    core::fmt::LowerExp::fmt(&x1, f)?;
    f.write_str(", ")?;
    core::fmt::LowerExp::fmt(&y1, f)?;
    f.write_str(", ")?;
    core::fmt::LowerExp::fmt(&z1, f)?;
    f.write_str(", ")?;
    core::fmt::LowerExp::fmt(&w1, f)?;
    f.write_str(" |\n| ")?;
    core::fmt::LowerExp::fmt(&x2, f)?;
    f.write_str(", ")?;
    core::fmt::LowerExp::fmt(&y2, f)?;
    f.write_str(", ")?;
    core::fmt::LowerExp::fmt(&z2, f)?;
    f.write_str(", ")?;
    core::fmt::LowerExp::fmt(&w2, f)?;
    f.write_str(" |\n| ")?;
    core::fmt::LowerExp::fmt(&x3, f)?;
    f.write_str(", ")?;
    core::fmt::LowerExp::fmt(&y3, f)?;
    f.write_str(", ")?;
    core::fmt::LowerExp::fmt(&z3, f)?;
    f.write_str(", ")?;
    core::fmt::LowerExp::fmt(&w3, f)?;
    f.write_str(" |")
  }
}

impl core::fmt::UpperExp for Mat4 {
  /// UpperExp formats like Display, but with the upper exponent.
  ///
  /// Passes the formatter along to the fields, so you can use any normal `f32`
  /// UpperExp format arguments that you like.
  fn fmt(&self, f: &mut core::fmt::Formatter) -> core::fmt::Result {
    let [x0, x1, x2, x3]: [f32; 4] = cast(self.x_axis);
    let [y0, y1, y2, y3]: [f32; 4] = cast(self.y_axis);
    let [z0, z1, z2, z3]: [f32; 4] = cast(self.z_axis);
    let [w0, w1, w2, w3]: [f32; 4] = cast(self.w_axis);
    //
    f.write_str("| ")?;
    core::fmt::UpperExp::fmt(&x0, f)?;
    f.write_str(", ")?;
    core::fmt::UpperExp::fmt(&y0, f)?;
    f.write_str(", ")?;
    core::fmt::UpperExp::fmt(&z0, f)?;
    f.write_str(", ")?;
    core::fmt::UpperExp::fmt(&w0, f)?;
    f.write_str(" |\n| ")?;
    core::fmt::UpperExp::fmt(&x1, f)?;
    f.write_str(", ")?;
    core::fmt::UpperExp::fmt(&y1, f)?;
    f.write_str(", ")?;
    core::fmt::UpperExp::fmt(&z1, f)?;
    f.write_str(", ")?;
    core::fmt::UpperExp::fmt(&w1, f)?;
    f.write_str(" |\n| ")?;
    core::fmt::UpperExp::fmt(&x2, f)?;
    f.write_str(", ")?;
    core::fmt::UpperExp::fmt(&y2, f)?;
    f.write_str(", ")?;
    core::fmt::UpperExp::fmt(&z2, f)?;
    f.write_str(", ")?;
    core::fmt::UpperExp::fmt(&w2, f)?;
    f.write_str(" |\n| ")?;
    core::fmt::UpperExp::fmt(&x3, f)?;
    f.write_str(", ")?;
    core::fmt::UpperExp::fmt(&y3, f)?;
    f.write_str(", ")?;
    core::fmt::UpperExp::fmt(&z3, f)?;
    f.write_str(", ")?;
    core::fmt::UpperExp::fmt(&w3, f)?;
    f.write_str(" |")
  }
}

impl Index<usize> for Mat4 {
  type Output = Vec4;
  #[inline(always)]
  fn index(&self, index: usize) -> &Vec4 {
    match index {
      0 => &self.x_axis,
      1 => &self.y_axis,
      2 => &self.z_axis,
      3 => &self.w_axis,
      otherwise => panic!("Mat4 index out of bounds: {}", otherwise),
    }
  }
}
impl IndexMut<usize> for Mat4 {
  #[inline(always)]
  fn index_mut(&mut self, index: usize) -> &mut Vec4 {
    match index {
      0 => &mut self.x_axis,
      1 => &mut self.y_axis,
      2 => &mut self.z_axis,
      3 => &mut self.w_axis,
      otherwise => panic!("Mat4 index out of bounds: {}", otherwise),
    }
  }
}

impl AsRef<[Vec4; 4]> for Mat4 {
  #[inline(always)]
  fn as_ref(&self) -> &[Vec4; 4] {
    cast_ref(self)
  }
}
impl AsMut<[Vec4; 4]> for Mat4 {
  #[inline(always)]
  fn as_mut(&mut self) -> &mut [Vec4; 4] {
    cast_mut(self)
  }
}

impl From<[f32; 16]> for Mat4 {
  #[inline]
  fn from([xx, xy, xz, xw, yx, yy, yz, yw, zx, zy, zz, zw, wx, wy, wz, ww]: [f32; 16]) -> Self {
    Self {
      x_axis: Vec4::from([xx, xy, xz, xw]),
      y_axis: Vec4::from([yx, yy, yz, yw]),
      z_axis: Vec4::from([zx, zy, zz, zw]),
      w_axis: Vec4::from([wx, wy, wz, ww]),
    }
  }
}
impl From<Mat4> for [f32; 16] {
  #[inline]
  fn from(
    Mat4 {
      x_axis,
      y_axis,
      z_axis,
      w_axis,
    }: Mat4,
  ) -> Self {
    let [e0, e1, e2, e3]: [f32; 4] = x_axis.into();
    let [e4, e5, e6, e7]: [f32; 4] = y_axis.into();
    let [e8, e9, e10, e11]: [f32; 4] = z_axis.into();
    let [e12, e13, e14, e15]: [f32; 4] = w_axis.into();
    [
      e0, e1, e2, e3, e4, e5, e6, e7, e8, e9, e10, e11, e12, e13, e14, e15,
    ]
  }
}

#[cfg(feature = "mint")]
impl From<mint::ColumnMatrix4<f32>> for Mat4 {
  #[inline]
  fn from(mint::ColumnMatrix4 { x, y, z, w }: mint::ColumnMatrix4<f32>) -> Self {
    let x_axis: Vec4 = x.into();
    let y_axis: Vec4 = y.into();
    let z_axis: Vec4 = z.into();
    let w_axis: Vec4 = w.into();
    Self {
      x_axis,
      y_axis,
      z_axis,
      w_axis,
    }
  }
}
#[cfg(feature = "mint")]
impl From<Mat4> for mint::ColumnMatrix4<f32> {
  #[inline]
  fn from(
    Mat4 {
      x_axis,
      y_axis,
      z_axis,
      w_axis,
    }: Mat4,
  ) -> Self {
    let x: mint::Vector4<f32> = x_axis.into();
    let y: mint::Vector4<f32> = y_axis.into();
    let z: mint::Vector4<f32> = z_axis.into();
    let w: mint::Vector4<f32> = w_axis.into();
    Self { x, y, z, w }
  }
}

#[cfg(feature = "serde")]
impl serde::Serialize for Mat4 {
  #[inline]
  fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
  where
    S: serde::Serializer,
  {
    <[f32; 16]>::from(*self).serialize(serializer)
  }
}
#[cfg(feature = "serde")]
impl<'de> serde::Deserialize<'de> for Mat4 {
  #[inline]
  fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
  where
    D: serde::Deserializer<'de>,
  {
    Ok(Self::from(<[f32; 16]>::deserialize(deserializer)?))
  }
}

impl Add for Mat4 {
  type Output = Self;
  #[inline]
  fn add(self, rhs: Self) -> Self {
    Self {
      x_axis: self.x_axis + rhs.x_axis,
      y_axis: self.y_axis + rhs.y_axis,
      z_axis: self.z_axis + rhs.z_axis,
      w_axis: self.w_axis + rhs.w_axis,
    }
  }
}
impl Add<f32> for Mat4 {
  type Output = Self;
  #[inline]
  fn add(self, rhs: f32) -> Self {
    self + Self::splat(rhs)
  }
}
impl Add<Mat4> for f32 {
  type Output = Mat4;
  #[inline]
  fn add(self, rhs: Mat4) -> Mat4 {
    Mat4::splat(self) + rhs
  }
}
impl AddAssign for Mat4 {
  #[inline]
  fn add_assign(&mut self, rhs: Self) {
    *self = *self + rhs
  }
}
impl AddAssign<f32> for Mat4 {
  #[inline]
  fn add_assign(&mut self, rhs: f32) {
    *self = *self + Self::splat(rhs)
  }
}

impl Sub for Mat4 {
  type Output = Self;
  #[inline]
  fn sub(self, rhs: Self) -> Self {
    Self {
      x_axis: self.x_axis - rhs.x_axis,
      y_axis: self.y_axis - rhs.y_axis,
      z_axis: self.z_axis - rhs.z_axis,
      w_axis: self.w_axis - rhs.w_axis,
    }
  }
}
impl Sub<f32> for Mat4 {
  type Output = Self;
  #[inline]
  fn sub(self, rhs: f32) -> Self {
    self - Self::splat(rhs)
  }
}
impl Sub<Mat4> for f32 {
  type Output = Mat4;
  #[inline]
  fn sub(self, rhs: Mat4) -> Mat4 {
    Mat4::splat(self) - rhs
  }
}
impl SubAssign for Mat4 {
  #[inline]
  fn sub_assign(&mut self, rhs: Self) {
    *self = *self - rhs
  }
}
impl SubAssign<f32> for Mat4 {
  #[inline]
  fn sub_assign(&mut self, rhs: f32) {
    *self = *self - Self::splat(rhs)
  }
}

impl Neg for Mat4 {
  type Output = Self;
  #[inline]
  fn neg(self) -> Self {
    Self {
      x_axis: -self.x_axis,
      y_axis: -self.y_axis,
      z_axis: -self.z_axis,
      w_axis: -self.w_axis,
    }
  }
}

impl Mul<f32> for Mat4 {
  type Output = Self;
  #[inline]
  fn mul(self, rhs: f32) -> Self {
    Self {
      x_axis: self.x_axis * rhs,
      y_axis: self.y_axis * rhs,
      z_axis: self.z_axis * rhs,
      w_axis: self.w_axis * rhs,
    }
  }
}
impl Mul<Mat4> for f32 {
  type Output = Mat4;
  #[inline]
  fn mul(self, rhs: Mat4) -> Mat4 {
    rhs * self
  }
}
impl MulAssign<f32> for Mat4 {
  #[inline]
  fn mul_assign(&mut self, rhs: f32) {
    *self = *self * rhs;
  }
}

impl Mul<Vec4> for Mat4 {
  type Output = Vec4;
  #[inline]
  fn mul(self, v4: Vec4) -> Vec4 {
    Vec4::new(
      Vec4::new(
        self.x_axis.x(),
        self.y_axis.x(),
        self.z_axis.x(),
        self.w_axis.x(),
      )
      .dot(v4),
      Vec4::new(
        self.x_axis.y(),
        self.y_axis.y(),
        self.z_axis.y(),
        self.w_axis.y(),
      )
      .dot(v4),
      Vec4::new(
        self.x_axis.z(),
        self.y_axis.z(),
        self.z_axis.z(),
        self.w_axis.z(),
      )
      .dot(v4),
      Vec4::new(
        self.x_axis.w(),
        self.y_axis.w(),
        self.z_axis.w(),
        self.w_axis.w(),
      )
      .dot(v4),
    )
  }
}

impl Mul for Mat4 {
  type Output = Self;
  #[inline]
  fn mul(self, rhs: Self) -> Self {
    Self {
      x_axis: self * rhs.x_axis,
      y_axis: self * rhs.y_axis,
      z_axis: self * rhs.z_axis,
      w_axis: self * rhs.w_axis,
    }
  }
}
impl MulAssign for Mat4 {
  #[inline]
  fn mul_assign(&mut self, rhs: Self) {
    *self = *self * rhs;
  }
}

/// ## Accessors
impl Mat4 {
  /// Obtains the `x_axis` column of this matrix
  #[inline(always)]
  pub fn x_axis(self) -> Vec4 {
    self.x_axis
  }

  /// Obtains the `y_axis` column of this matrix
  #[inline(always)]
  pub fn y_axis(self) -> Vec4 {
    self.y_axis
  }

  /// Obtains the `z_axis` column of this matrix
  #[inline(always)]
  pub fn z_axis(self) -> Vec4 {
    self.z_axis
  }

  /// Obtains the `w_axis` column of this matrix
  #[inline(always)]
  pub fn w_axis(self) -> Vec4 {
    self.w_axis
  }

  /// `&mut` to the `x_axis` column of this matrix
  #[inline(always)]
  pub fn x_axis_mut(&mut self) -> &mut Vec4 {
    &mut self.x_axis
  }

  /// `&mut` to the `y_axis` column of this matrix
  #[inline(always)]
  pub fn y_axis_mut(&mut self) -> &mut Vec4 {
    &mut self.y_axis
  }

  /// `&mut` to the `z_axis` column of this matrix
  #[inline(always)]
  pub fn z_axis_mut(&mut self) -> &mut Vec4 {
    &mut self.z_axis
  }

  /// `&mut` to the `w_axis` column of this matrix
  #[inline(always)]
  pub fn w_axis_mut(&mut self) -> &mut Vec4 {
    &mut self.w_axis
  }
}

/// ## Constructors
impl Mat4 {
  /// Combines the columns given into a new `Mat4`
  #[inline(always)]
  pub fn new(x_axis: Vec4, y_axis: Vec4, z_axis: Vec4, w_axis: Vec4) -> Self {
    Self {
      x_axis,
      y_axis,
      z_axis,
      w_axis,
    }
  }

  /// Splats the given value across all columns.
  #[inline]
  pub fn splat(v: f32) -> Self {
    Self {
      x_axis: Vec4::splat(v),
      y_axis: Vec4::splat(v),
      z_axis: Vec4::splat(v),
      w_axis: Vec4::splat(v),
    }
  }

  /// Forms a diagonal matrix from the values given.
  #[inline]
  pub fn diagonal(x: f32, y: f32, z: f32, w: f32) -> Self {
    let x_axis = Vec4::new(x, 0.0, 0.0, 0.0);
    let y_axis = Vec4::new(0.0, y, 0.0, 0.0);
    let z_axis = Vec4::new(0.0, 0.0, z, 0.0);
    let w_axis = Vec4::new(0.0, 0.0, 0.0, w);
    Self {
      x_axis,
      y_axis,
      z_axis,
      w_axis,
    }
  }

  /// Down-converts to a [`Mat3`]
  #[inline]
  pub fn to_mat3(self) -> Mat3 {
    let x_axis = self.x_axis.to_vec3();
    let y_axis = self.y_axis.to_vec3();
    let z_axis = self.z_axis.to_vec3();
    Mat3 {
      x_axis,
      y_axis,
      z_axis,
    }
  }
}
